/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
 * Copyright (C) 1992 1993 1994 1995 1996 1997 1998 Michael Kraehe, Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * lib/maus.c: Allgemeine Hilfsfunktionen fuer den MausTausch.
 */

#include "mausconsys.h"

#include <stdio.h>
#include <syslog.h>
#include <assert.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "maus.h"
#include "domains.h"

#include "msg_io.h"

#include "groups.h"
#include "uoio.h"

#define LINEEND "\r\n"

static void maus_mk_ref (metamsg_t * msg, uostr_t *mref, uostr_t *realref);
static size_t export_part (FILE * f, metamsg_t * meta, long partnr);

#if 0
static void add_ref (metalist_t ** r, char *s);
#endif

	/*
	 * Eine Adresse im MausTauschformat (Uwe Ohse @ DU3) verwandeln in
	 * RFC-konforme Adress- und Realnameteile, Uwe_Ohse@DU3.maus.de und
	 * "Uwe Ohse".
	 */
static BOOLEAN mausadr_to_rfcadr (
				 char *mausadr,		/* zu wandelnde Adresse */
				 uostr_t *adresse,		/* entstehende Adresspart */
				 uostr_t *realname /* Realname, nur ber den Localpart */ );

/*
 * Eine Adresse im MausTauschformat ausgeben.
 * bernimmt die Rckwandlung von domainisierten Boxnamen 
 * (mausdu3.gun.de -> DU3) [notwendig weil MAUS zu dumm ist das selber
 * zu machen] und Fidoadressen [notwendig weil FIDO halt FIDO ist].
 */
struct mt_private {
	BOOLEAN mayclose;
	BOOLEAN output;
	BOOLEAN eof;
	char *fname;
	uoio_t uoio; /* input */
	FILE *f; /* output */

	BOOLEAN zurueckgestellt;
	char *line;
	size_t llen;
};

static mio_retcode_t
 mt_open (const char *host, const char *mode,
		  mio_t ** ret_handle, uostr_t *errbuf);
static mio_retcode_t mt_close (mio_t * chan);
static mio_retcode_t mt_export (mio_t * chan, metamsg_t *);
static mio_retcode_t mt_import (mio_t * chan, metamsg_t *);

mio_base_t mio_maustausch =
{
	"maustausch",

	mt_open,
	mt_import,
	mt_export,
	mt_close,

	0							/* usecount */
};


static mio_retcode_t
mt_open (const char *fname, const char *mode,
		 mio_t ** ret_handle, uostr_t *errbuf)
{
	BOOLEAN output = FALSE;
	struct mt_private *priv;
	mio_t *neu;

	if (!strcmp (mode, "w"))
		output = TRUE;
	else if (strcmp (mode, "r")) {
		uostr_xdup_cstrmulti(errbuf,"illegal mode ",mode," for mt_open", NULL);
		uostr_x0(errbuf);
		log (LOG_ERR, "Illegal mode %s for mt_open", mode);
		return MIO_ERROR;
	}
	neu = mio_alloc (&mio_maustausch);
	priv = malloc (sizeof (struct mt_private));
	if (priv)
		priv->fname = strdup (fname);
	if (!neu || !priv || !priv->fname) {
		free (neu);
		if (priv)
			free (priv->fname);
		free (priv);
		uostr_xdup_cstr(errbuf,_("out of memory"));
		uostr_x0(errbuf);
		log (LOG_ERR, _ ("out of memory"));
		return MIO_RETRY;
	}
	priv->zurueckgestellt = FALSE;
	priv->eof = FALSE;
	priv->line = NULL;
	priv->output = output;
	priv->uoio.buf=priv->uoio.rstart=NULL;
	if (!strcmp (fname, "-")) {
		if (output)
			priv->f = stdout;
		else {
			priv->f = NULL;
			uoio_assign_r(&priv->uoio,0,read,0);
		}
		priv->mayclose = FALSE;
	} else {
		int fail=0;
		if (output) {
			priv->f = fopen (fname, "w");
			if (!priv->f)
				fail=1;
		} else {
			int fd;
			fd=open(fname,O_RDONLY);
			if (fd==-1) {
				fail=1;
			} else {
				priv->f = NULL;
				uoio_assign_r(&priv->uoio,fd,read,0);
			}
		}
		if (fail) {
			int errno1 = errno;
			free (priv->fname);
			free (priv);
			free (neu);
			uostr_xdup_cstrmulti(errbuf,"cannot open ",fname,": ", strerror (errno1),NULL);
			uostr_x0(errbuf);
			log (LOG_ERR, _ ("cannot open %s: %s"), fname, strerror (errno1));
			return MIO_RETRY;
		}
		priv->mayclose = TRUE;
	}
	neu->priv_data = priv;
	*ret_handle = neu;
	return MIO_OK;
}

static mio_retcode_t
mt_close (mio_t * chan)
{
	mio_retcode_t ok = MIO_OK;
	int errno1 = 0;
	struct mt_private *priv;

	assert (chan);

	priv = (struct mt_private *) chan->priv_data;

	assert (priv);

	if (!priv)
		return TRUE;
	if (priv->mayclose && priv->f) {
		if (fclose (priv->f) != 0) {
			/* wir haben keine Mglichkeit, den Fehler an dieser Stelle
			 * nach oben zu melden */
			errno1 = errno;
			log (LOG_ERR, _ ("cannot close %s: %s"), priv->fname, strerror (errno1));
			ok = MIO_ERROR;
		}
	}
	free (priv->fname);
	uoio_destroy(&priv->uoio);
	free (priv);
	chan->priv_data = NULL;
	mio_free (chan);
	errno = errno1;
	return ok;
}

BOOLEAN
maus_print_adr (FILE * f, const char *prefix, char *adr, const char *postfix)
{
	box_t *box;
	char *at;
	char *domainpart;
	static uostr_t localpart=UOSTR_INIT; /* memory buffer, not really uostr */
	size_t l;
	if (!postfix) postfix="";
	if (!prefix) prefix="";

	l = strlen (adr);
	uostr_xneed(&localpart,l+1);
	if (opts.workmode == USER) {
		char *q;
		char *s;
		for (q = adr, s = localpart.data; *q && *q != '@'; q++) {
			if (*q == '\\') {
				q++;
				*s++ = *q;
			} else if (*q == '_')
				*s++ = ' ';
			else
				*s++ = *q;
		}
		*s = 0;
		if (*q == '@')
			at = q;
		else
			at = NULL;
	} else {
		at = strchr (adr, '@');
		if (!at)
			memcpy (localpart.data, adr, l + 1);
		else {
			memcpy (localpart.data, adr, at - adr);
			localpart.data[at - adr] = 0;
		}
	}

	if (!at) {
		if (0 > fprintf (f, "%s%s%s", prefix, localpart.data, postfix))
			return FALSE;
		return TRUE;
	}
	domainpart = at + 1;
	while (isspace (*domainpart))	/* yes, this should not happen */
		domainpart++;
	l = strlen (domainpart);

	box = dom_find_domain (domainpart);
	if (box && box->kuerzel) {
		/* OK, dann auch mal den Localpart bersetzen */
		unsigned char *lp;
		lp=localpart.data;
		while (isalnum(*lp) || isspace(*lp) || *lp=='_')
			lp++;
		if (!*lp) {
			for (lp=localpart.data;*lp;lp++) {
				if (*lp=='_')
					*lp=' ';
			}
		}
		if (opts.strip_domain) {
			if (0 > fprintf (f, "%s%s@%s%s", prefix, localpart.data, box->kuerzel, postfix))
				return FALSE;
			return TRUE;
		}
	}

  	if (0 > fprintf (f, "%s%s@%s%s", prefix, localpart.data, domainpart, postfix))
		return FALSE;
	return TRUE;
}

static BOOLEAN
mausadr_to_rfcadr (char *mausadr, uostr_t *adress, uostr_t *realname)
{
	uostr_t domain=UOSTR_INIT;
	char *p,*q;
	box_t *box;
	BOOLEAN ok = TRUE;

	/* fr saubere Verhltnisse sorgen */
	uostr_cut(adress,0);
	uostr_cut(realname,0);

	/* Erster Teil: In Local- und Domainpart zerlegen */
	for (p=mausadr;*p && *p!='@';p++)
		/* nothing */;
	uostr_xneed(realname,(size_t)(p-mausadr+1));
	p=mausadr;
	q=realname->data;
	while (isspace(*p))
		p++;
	while (*p!='@' && *p) {
		/* mehrere Leerzeichen zusammenziehen */
		if ((p[0]==' '||p[0]=='\t') && (p[1]==' '||p[1]=='\t'))
			p++;
		else
			*q++=*p++;
	}
	while (q>realname->data) {
		if (q[-1]==' '||q[-1]=='\t')
			q--;
		else
			break;
	}
	realname->len=q-realname->data;
	uostr_x0(realname);

	if (*p=='@') {
		size_t l=strlen(p); /* includes 1 char "too much" for \0 */
		uostr_xneed((&domain),l);
		q=domain.data;
		p++;
		while (isspace(*p))
			p++;
		while (*p) {
			/* mehrere Leerzeichen zusammenziehen */
			if (p[0]==' ' && p[1]==' ')	
				p++;
			else
				*q++=*p++;
		}
		while (q>domain.data) {
			if (q[-1]==' '||q[-1]=='\t')
				q--;
			else
				break;
		}
		domain.len=q-domain.data;
		uostr_x0(&domain);
	}

	/* 
	 * So, jetzt haben wir zwei Strings darin. Der erste ist der Localpart,
	 * mit Leerzeichen darin. Der zweite der Domainpart, mglicherweise 
	 * ebenfalls mit Leerzeichen darin, was auf Fido-hnliche Adressierung
	 * hindeutet.
	 */

	/* Wenn es ein Boxkrzel ist */
	if (domain.data && !strpbrk (domain.data, ".\t ")) {
		char *kopie = xentumlauten1 (domain.data);
		box = dom_find_boxkuerzel (kopie);
		free (kopie);
		if (box) {
			if (box->primdomain) {
				uostr_xdup_cstr(&domain,box->primdomain);
			} else {
				uostr_cut(&domain,-1);
				uostr_xadd_cstr(&domain,opts.mausdomain);
			}
			uostr_x0(&domain);
		}
	}

#if 0
	/*
	 * In jedem Fall wird der Domainpart entumlautet.
	 */
	convert_umlaute (domain);
#endif

	{	
		/* 
		 * * So, jetzt haben wir eine gltige Domain.
		 * * Jetzt die Adresse zusammensetzen ...
		 */
		char *p1, *p2;
		p1 = xentumlauten1 (realname->data);
		if (domain.data) {
			p2 = xentumlauten1 (domain.data);
			uostr_xdup_cstrmulti(adress,p1,"@",p2,NULL);
			free (p2);
		} else 
			uostr_xdup_cstr(adress,p1);
		uostr_x0(adress);
		free (p1);
	}
	/* und Leerzeichen durch underscores ersetzen ... */
	p = adress->data;
	while (*p) {
		if (*p == ' ')
			*p = '_';
		p++;
	}

#if 0
	/* Eventuell fehlende Domain ersetzen */
	if (!strchr (domain, '.') && *domain)
		strcpy (adress + strlen (adress), opts.mausdomain);
#endif

	/* Ein Username ohne Spaces knnte angegeben worden sein, in dem
	 * Fall gibt es keinen Realnamen 
	 */
	if (!strchr (realname->data, ' '))
		uostr_cut(realname,0);
	uostr_freedata(&domain);
	return ok;
}


static char *
last_reference (metalist_t * r)
{
	if (!r)
		return NULL;
	while (r->next)
		r = r->next;
	return r->text;
}


static size_t
write_message_refs (FILE * f, metamsg_t * meta)
{
	char *ref;

	ref = last_reference (meta->references);
	if (!ref || !strchr (ref, '@'))
		return 0;

	if (!opts.send_short_ids)
		fprintf (f, "-%s" LINEEND, ref);
	else {
		/* Fr Software die seit 1990 verspricht bald lange IDs 
		 * zu knnen.
		 */
		/* static weil diese Routine oft aufgerufen wird ... */
		static uostr_t mref=UOSTR_INIT;
		static uostr_t rref=UOSTR_INIT;
		uostr_fcut(&mref);
		uostr_fcut(&rref);
		maus_mk_ref (meta, &mref, &rref);
		if (mref.len && mref.data[0]) {
			/* 
			 * Eine mausinterne References konnte ermittelt werden 
			 * ### HACK ALARM: Abhngigkeit von A/P!
			 */
			if ((mref.data[0] == 'A' && !meta->is_mail)
				|| (mref.data[0] == 'P' && meta->is_mail)
				|| (mref.data[0] != 'P' && mref.data[0] != 'A'))
				fprintf (f, "-%s" LINEEND, mref.data);
		}
		if (rref.len && rref.data[0]) {
			/* Und eine echte ebenso */
			fprintf (f, "R%s" LINEEND, rref.data);
		}
	}
	return 0;					/* Kein Bodyplatz belegt */
}

static void
write_message_ids (FILE * f, metamsg_t * meta, long partnr)
{
	if (opts.workmode == USER) {
		/* 
		 * Als Benutzer sind keine echten Message-IDs erlaubt, man mu
		 * temporre IDs vergeben. Die Mausprogrammierer halten das 
		 * auch noch fr sinnvoll, angeblich wre Chaos die Folge,
		 * drften Benutzer IDs selbst vergeben. Aber welchen Sinn 
		 * hat es, eine *eindeutige* Message-ID (und die erzeugen 
		 * Mail- und Newssysteme nunmal) wegzuwerfen, um eine
		 * neue, weniger eindeutige, zu erzeugen?
		 * IMO eine vllig berzogene Rcksichtnahme auf veraltete
		 * Systeme, zumindest fr die sicheren Systeme sollte da
		 * mehr erlaubt sein.
		 */
		/* Message-ID */
		static char *last_id=NULL;
		static char *freeme=NULL;
		static int lfd=0;
		char *t;
		if (freeme) {
			free(freeme);
			freeme=NULL;
		}
		if (!last_id) {
			last_id=XSTRDUP(meta->transfer_id);
			t=last_id;
		}
		else if (0==strcmp(meta->transfer_id,last_id)) {
			/* uhh ... */
			char num[STR_ULONG];
			str_ulong(num,++lfd);
			freeme=xconcat(meta->transfer_id,".",num,NULL);
			t=freeme;
		} else {
			free(last_id);
			last_id=XSTRDUP(meta->transfer_id);
			t=last_id;
		}
		if (partnr)
			fprintf (f, "#%s%ld" LINEEND, t, partnr);
		else
			fprintf (f, "#%s" LINEEND, t);

	}
	/* Usermodus? */
	else {
		char *idptr;
		/* t-online ist so schlecht ... es sendet Kommentare mit
		 * derselben ID zurck. Schliet die Gates! */
		if (meta->is_mail && opts.t_online_id_hack) {
			if (((meta->env_from && strstr (meta->env_from, "t-online.de"))
				 || (meta->head_from && strstr (meta->head_from->text, "t-online.de")))
				&& !strstr (meta->message_id, "t-online.de")) {
				char *s;
				s = concat (meta->message_id, ".t-online.hack", NULL);
				if (s) {
					/* kann das nicht ins MausNet senden wegen 16K-Grenzen-Problem und
					 * der Unfhigkeit, unbekannte Header durchzureichen. Bug! */
					meta_add_trace(meta,MMT_NX|MMT_ID,
							   "id hacked (t-online-hack: %s)", meta->message_id);
					free (meta->message_id);
					meta->message_id = s;
				}				/* else hope for the best */
			}
		}
		if (partnr) {
			size_t l;
			char numbuf[STR_ULONG];
			static uostr_t idbuf=UOSTR_INIT;
			/* 
			 * Splitten nach Splitstandard.
			 * 1. Teil: localpart@domain
			 * 2. Teil: localpart@domain:2
			 * n. Teil: localpart@domain:n
			 */
			l=strlen(meta->message_id);
			uostr_xneed(&idbuf,l+1); /* : */
			uostr_xdup_mem(&idbuf,meta->message_id,l);
			idbuf.data[idbuf.len++]=':';
			l=str_ulong(numbuf,partnr);
			uostr_xadd_mem(&idbuf,numbuf,l+1); /* incl. \0 */
			idptr=idbuf.data;
		}
		else
			idptr=meta->message_id;
		if (!opts.send_short_ids) {
			/* die Box will lange IDs */
			/*
			 * keine transfer-id -> wir arbeiten als Filter oder so.
			 * transparent-IDs: als einfacher konverter arbeiten.
			 */
			if (meta->transfer_id && !opts.transparent_ids) {
				fprintf (f, "#%s" LINEEND, meta->transfer_id);
				fprintf (f, "I%s" LINEEND, idptr);
			} else
				fprintf (f, "#%s" LINEEND, idptr);
		} else {
			/* Box will kurze IDs */
			/* maus braucht uniq transferid. Und wir mssen die richtige
			 * wenigstens einmal fr maus2rfc haben ... zum Lschen der
			 * News nmlich. 
			 */
			static char *last_id=NULL;
			static int lfd=0;
			char *m=maus_long_to_short_id(meta->message_id);
			/* Bei PMs macht MAUS einen Dupecheck - auf die ID. Auch bei
			 * Nachrichten von Listen. Ergo: _nicht_ die originale ID
			 * restaurieren (!meta->is_mail)
			 * Auerdem mssen bei ffentlichen Nachrichten die IDs mit
			 * A anfangen 
			 */
			if (m && !meta->is_mail && *m=='A') {
				/* wir haben aus der langen die kurze ID regenerieren knnen */
				/* nun hoffen wir mal da da nur ein Teil kommt! */
				/* XXX damit verlieren wir auch die Zuordnung im #LOG. Nicht so gut,
				 * XXX aber was tun? Batch umbenennen? Von hier aus unmglich ...
				 */
				fprintf (f, "#%s" LINEEND , m);
				free(m);
			} else {
				if (!last_id) {
					last_id=XSTRDUP(meta->transfer_id);
					fprintf (f, "#%s" , meta->transfer_id);
				} else if (0!=strcmp(last_id,meta->transfer_id)) {
					free(last_id);
					last_id=XSTRDUP(meta->transfer_id);
					lfd=0;
					fprintf (f, "#%s" , meta->transfer_id);
				} else {
					lfd++;
					fprintf (f, "#%s.%d" , meta->transfer_id,lfd);
				}
				if (partnr)
					fprintf (f, ":%ld" LINEEND, partnr);
				else
					fprintf (f, "%s",LINEEND);
			}
			fprintf (f, "I%s" LINEEND, idptr);
		}
	}							/* User- oder Gatemodus */
}

static size_t
export_part (FILE * f, metamsg_t * meta, long partnr)
{

	int is_crossposting = 0;
	int is_mail = 0;
	int is_mime = 0;
	size_t textsize = 0;
	int used_cludges = 0;
	char *ref;
	char *m_content_type_for_body = NULL;
	BOOLEAN header_envelope_mismatch = FALSE;
	BOOLEAN need_pseudoheader=FALSE;
	BOOLEAN need_followup_to_kludge=FALSE;

	if (meta->references)
		ref = last_reference (meta->references);
	else
		ref = NULL;

	/* Einen Header plus eventuell Pseudoheader in die Datei schreiben */

	/*
	 * Die Message-ID schreiben.
	 */
	write_message_ids (f, meta, partnr);

	if (meta->is_mail) {
		/* Dann ist es eine Mail */
		char *primary = NULL;
		char *secondary = NULL;

		if (meta->env_to)
			primary = meta->env_to;
		if (meta->head_to) {
			if (!primary)
				primary = meta->head_to->text;
			else
				secondary = meta->head_to->text;
		}
		is_mail = 1;
		if (opts.send_b_lines)
			fputs("BN" LINEEND, f);

		if (primary) {			/* Anderes kommt hier nicht an */
			maus_print_adr (f, "A", primary, LINEEND);
			/* nun prfen wir, ob sich envelope- und Headerempfnger 
			 * deutlich unterscheiden. Falls ja, wird spter ein
			 * entsprecheder Vermerk in den Body gesetzt. Das dient
			 * der Entschrfung der Mailinglistenproblematik.
			 */
			if (secondary) {
				/* belster Code von Welt ... */
				char *p1 = strchr (primary, '@');
				char *p2 = strchr (secondary, '@');
				char *q1, *q2;
				size_t l1;
				size_t l2;

				/* Nein, das ist *nicht nur* Paranoia. Ich hab' schon 
				 * Adressen mit Spaces drin gesehen. Nicht da die 
				 * korrekt gewesen wren, aber aufgetreten sind sie 
				 */
				if (p1)
					q1 = p1 - 1;
				else
					q1 = primary + strlen (primary) - 1;
				while (isspace (*q1))
					q1--;
				p1 = primary;
				while (isspace (*p1))
					p1++;
				l1 = q1 - p1;

				if (p2)
					q2 = p2 - 1;
				else
					q2 = secondary + strlen (secondary) - 1;
				while (isspace (*q2))
					q2--;
				p2 = secondary;
				while (isspace (*p2))
					p2++;
				l2 = q2 - p2;

				if (l1) {
					if (l2 != l1)
						header_envelope_mismatch = TRUE;
					else if (0 != memcmp (p1, p2, l1))
						header_envelope_mismatch = TRUE;
					if (header_envelope_mismatch)
						need_pseudoheader=TRUE;
				}
			}
		}
		write_message_refs (f, meta);
	} else if (!meta->newsgroups) {
		/* ### LOG-Eintrag fehlt */
		return (size_t) - 1;
	} else {
		/* ffentliche Nachricht */
		metalist_t *ng;
		BOOLEAN grp_found = FALSE;
		int need_warning=0;
		int numgroups=0;
		int printgroups=0;

		/* Eine Gruppe kann auf jeden Fall gewandelt werden, das ist
		 * weiter oben schon geprft worden.
		 * in mindestens eine Gruppe kann auch importiert werden! siehe
		 * rfc2maus.c.
		 */
		/* Logik: 
		 * wenn crosspostings erlaubt sind: schreibe jede Gruppe hinein.
		 * sonst:
		 *   gehe durch die Liste bis zur ersten bekannten Gruppe. Schreibe
		 *   sie hinein. Wenn keine Gruppe bekannt ist dann erste nehmen.
		 *   Wenn mehr als eine Gruppe dann merken da wir spter einen
		 *   Textklduge bentigen.
		 */
		for (ng=meta->newsgroups;ng;ng=ng->next) {
			grp_data_t *grp;
			numgroups++;
			grp = grp_suche_rfcname (ng->text, opts.pollsite);
			if (grp && grp->import_state==IMPORT_SKIP && !opts.send_crosspostings)
				continue;
			if (grp) {
				fprintf (f, "G%s" LINEEND, grp->mausname);
				printgroups++;
				grp_found = TRUE;
				if (!opts.send_crosspostings) {
					if (ng->next) {
						meta->mausflags.exportflag=XR_MINUS;
						need_warning++;
						ng=ng->next;
						while (ng) {
							ng=ng->next;
							numgroups++;
						}
					}
					break;
				}
			} else if (opts.send_crosspostings) {
				fprintf (f, "G%s" LINEEND, ng->text);
				printgroups++;
			} else if (numgroups > 1 || ng->next) {
				meta->mausflags.exportflag=XR_MINUS;
				need_warning++;
			}
		}
		if (need_warning)
			meta_add_trace(meta,MMT_NX|MMT_LOSS,"throwing %d groups away",numgroups-printgroups);
		if (meta->newsgroups && meta->newsgroups->next && !opts.send_crosspostings) {
			/* spter Textcludge erzeugen */
			is_crossposting = 1;
			need_pseudoheader=TRUE;
		}
		if (!grp_found && !opts.send_crosspostings) {
			/* Das kann zwar in rfc2maus *hier* nicht passieren, aber 
			 * mglicherweise in irgendwelchen Filtern, die die
			 * Gruppenliste nicht laden.
			 */
			fprintf (f, "G%s" LINEEND, meta->newsgroups->text);
		}
		write_message_refs (f, meta);

		if (meta->followup_to && opts.send_followup_to) {
			metalist_t *f_list_p;
			BOOLEAN found = FALSE;
			int total=0;
			int used=0;

			f_list_p = meta->followup_to;
			while (f_list_p) {
				total++;
				if (found && !opts.send_crosspostings) {
					f_list_p = f_list_p->next;
					continue;
				}
				if (!strcasecmp (f_list_p->text, "poster")) {
					fprintf (f, "Fposter" LINEEND);
					found=TRUE;
					used++;
				} else {
					grp_data_t *grp;
					grp = grp_suche_rfcname (f_list_p->text, opts.pollsite);
					if (grp) {
						fprintf (f, "F%s" LINEEND, grp->mausname);
						found = TRUE;
						used++;
					} else if (opts.send_crosspostings) {
						fprintf (f, "F%s" LINEEND, f_list_p->text);
						found = TRUE;
						used++;
					}
				}
				f_list_p = f_list_p->next;
			}
			/* Wenn berhaupt nichts bekannt war, dann eben so ... */
			if (!found) {
				used++;
				fprintf (f, "F%s" LINEEND, meta->followup_to->text);
			}
			if (used!=total) {
				meta_add_trace(meta,MMT_NX|MMT_LOSS,"throwing %d followup-groups away",
					total-used);
				meta->mausflags.exportflag=XR_MINUS;
				need_followup_to_kludge=TRUE;
				need_pseudoheader=TRUE;
			}
		} else if (meta->followup_to) {
			need_followup_to_kludge=TRUE;
			meta_add_trace(meta,MMT_NX|MMT_LOSS,"throwing followup-groups away");
			meta->mausflags.exportflag=XR_MINUS;
			need_pseudoheader=TRUE;
		}
	}

	/* User drfen keine Absenderangabe verschicken */
	if (opts.workmode != USER) {
		/* Absender angegeben? */
		/* Just for the case */
		char *from = NULL;
		if (meta->head_from) {
			from = meta->head_from->text;
			if (meta->head_from->ext)
				fprintf (f, "N%s" LINEEND, meta->head_from->ext);
		} else if (meta->env_from && meta->is_mail) {
			from = meta->env_from;
			if (meta->head_from && meta->head_from->text
				&& !strcasecmp (meta->head_from->text, meta->env_from)) {
				if (meta->head_from->ext)
					fprintf (f, "N%s" LINEEND, meta->head_from->ext);
			}
		} else if (meta->sender) {
			from = meta->sender->text;
			if (meta->sender->ext)
				fprintf (f, "N%s" LINEEND, meta->sender->ext);
		}
		if (from)
			maus_print_adr (f, "V", from, LINEEND);
	}
	/* Datumsangabe erzeugen */
	if (meta->date!=(uo_sec70_t) -1 && meta->date) {
		char timezwi[64];
		maus_make_date(timezwi,meta->date);
		fprintf (f, "E%s" LINEEND, timezwi);
	}
	/*
	 * Subject
	 */
	{
		size_t subjmax;			/* maximale Laenge des Subjects */
		char *p;

		if (!meta->subject)
			meta->subject=XSTRDUP("Kein Betreff angegeben");

		subjmax=opts.subjectlen;

		/* Re: eventuell killen */
		if (opts.rekillmode == REKILL_NONE)
			/* nicht killen */
			p = meta->subject;
		else if (opts.rekillmode == REKILL_FULL) {
			/* immer killen */
			if (!strncasecmp (meta->subject, "re: ", 4))
				p = meta->subject + 4;
			else
				p = meta->subject;
		} else {
			/* REKILL_SMART */
			/* nur killen, wenn das Maximum berschritten wrde */
			if (!strncasecmp (meta->subject, "re: ", 4)) {
				if (strlen (meta->subject) > subjmax)
					p = meta->subject + 4;
				else
					p = meta->subject;
			} else
				p = meta->subject;
		}
		if (strlen (p) > subjmax) {
			p[subjmax] = '\0';
			meta->mausflags.exportflag=XR_MINUS;
			meta_add_trace(meta,MMT_LOSS|MMT_NX,"shortening subject");
		}
		fprintf (f, "W%s" LINEEND, p);
	}


	if (opts.workmode != USER && meta->organization)
		fprintf (f, "O%s" LINEEND, meta->organization);

	if (opts.send_path && meta->path) {
		if (strlen (meta->path) > 250) {
			char *p;
			p = meta->path + 250;
			while (p > meta->path && *p != '!')
				p--;
			if (p > meta->path) {
				*p = '\0';
				fprintf (f, "p%s" LINEEND, meta->path);
				*p = '!';
			}
		} else
			fprintf (f, "p%s" LINEEND, meta->path);
	}
	if (meta->reply_to && opts.send_reply_to) {
		/* nur *ein* Reply-To, mehr ist wohl nicht sinnvoll, oder? */
		fprintf (f, "T%s" LINEEND, meta->reply_to->text);
		if (meta->reply_to->next) {
			meta->mausflags.exportflag=XR_MINUS;
			meta_add_trace(meta,MMT_LOSS|MMT_NX,"throwing excessive reply-to(s) away");
		}
	} else if (meta->reply_to) {
		/* we will send it as text kludge */
		meta->mausflags.exportflag=XR_MINUS;
		meta_add_trace(meta,MMT_MIME|MMT_NX,"put reply-to in MIME pseudoheader");
		need_pseudoheader=TRUE;
	}
	if (meta->sender && opts.send_sender) {
		fprintf (f, "S%s" LINEEND, meta->sender->text);
		if (meta->sender->next) {
			meta->mausflags.exportflag=XR_MINUS;
			meta_add_trace(meta,MMT_LOSS|MMT_NX,"throwing excessive sender(s) away");
		}
	} else if (meta->sender) {
		/* we will send it as text kludge */
		meta->mausflags.exportflag=XR_MINUS;
		meta_add_trace(meta,MMT_MIME|MMT_NX,"put sender in MIME pseudoheader");
		need_pseudoheader=TRUE;
	}

	/* einige Headerzeilen knnen wir nicht in MausTauschzeilen transportieren.
	 * Einige? Viele ... aber einige davon sind in jedem Falle wertvoll. So ...
	 */
	if (meta->keywords || meta->summary) {
		/* fr die erzeugen wir spter Zeilen im MIME-Pseudoheader */
		meta->mausflags.exportflag=XR_MINUS;
		meta_add_trace(meta,MMT_MIME|MMT_NX,"put extra headers in MIME pseudo header");
		need_pseudoheader=TRUE;
	}
	if (meta->content_id || meta->content_description)
		need_pseudoheader=TRUE;

	/* Feststellen, ob die Nachricht MIME ist */
	if (meta->mime_version) {
		char *v = rfc_getword (meta->mime_version, NULL);
		if (v && 0 == strcasecmp (v, "1.0"))
			is_mime = 1;
		free (v);
	} else if (need_pseudoheader) {
		meta->mime_version=XSTRDUP("1.0");
		is_mime=1;
	}
	if (!meta->content_type)
		meta->content_type = XSTRDUP ("text/plain");

	if (is_mime) {
		/* wir mssen jetzt einen MIME-Header generieren */
		/* Format:
		 * M1.0; text/enriched; charset=...
		 * Problem dabei: charset ist leider zur Zeit nicht machbar. 
		 */
		char *type;
		char *next;
		char *line = NULL;
		size_t linesize;
		char *wpos;
		size_t l;

		is_mime = 0;
		/* Typ holen. 1. Wort in der Zeile. */
		type = rfc_getword (meta->content_type, &next);
		if (!type)
			goto no_m_line;

		/* text/plain ohne Parameter behandeln wir gar nicht erst */
		/* wenn wir MIME aber fr Pseudoheader brauchen, nun gut ... */
		if (0 == strcasecmp (type, "text/plain") && !next && !need_pseudoheader) {
			free (type);
			goto no_m_line;
		}
		is_mime = 1;

		linesize = 1024;
		line = XMALLOC (linesize);

#define MIME_CONSTANT "M1.0; Content-Type: "
		memcpy (line, MIME_CONSTANT, sizeof (MIME_CONSTANT) - 1);
		wpos = line + sizeof (MIME_CONSTANT) - 1;
		l = strlen (type);
		if (l + (wpos - line) >= linesize) {
			size_t off = wpos - line;
			/* genug Platz schaffen */
			linesize = (linesize + l + 1024) & (~1024);
			line = XREALLOC (line, linesize);
			wpos = line + off;
		}
		memcpy (wpos, type, l);
		free (type);
		wpos += l;

		/* die folgende Schleife beseitigt unter anderem Kommentare */
		for (type = next ? rfc_getword (next, &next) : NULL;
			 type; type = next ? rfc_getword (next, &next) : NULL) {
			l = strlen (type);
			/* "; " == 2 */
			if (l + 2 + (wpos - line) >= linesize) {
				size_t off = wpos - line;
				/* genug Platz schaffen */
				linesize = (linesize + l + 2 + 1024) & (~1024);
				line = XREALLOC (line, linesize);
				wpos = line + off;
			}
			*wpos++ = ';';
			*wpos++ = ' ';
			memcpy (wpos, type, l);
			free (type);
			wpos += l;
		}
		*wpos = 0;
		if (wpos - line > 200) {
			/* Oh hilfe - das habe *ich* mit verschuldet (und ganz anders
			 * gemeint). */
			fputs ("M1.0; Content-Type: obscure/anything; "
				   "really too many parameters", f);

			/* "M1.0; " berspringen */
			m_content_type_for_body = XSTRDUP (line + 6);
			/* bis zum Beweis des Gegenteils glaube ich nicht da das gutgeht */
			meta->mausflags.exportflag=XR_MINUS;
			meta_add_trace(meta,MMT_NX|MMT_MIME,"too lang content-type line");
		} else
			fputs (line, f);
		fputs (LINEEND, f);
	  no_m_line:
		free (line);
	}
	if (opts.workmode!=USER) {
		/* XR_PLUS wird nie gesetzt, wir setzen nur XR_MINUS */
		/* XR_NOTSET == XR_PLUS? _hier_ ja, bei maus->rfc nicht! */
		if (meta->mausflags.exportflag==XR_MINUS && !meta->is_mail) {
			fputs ("XR-\r\n", f);
		} else {
			fputs ("XR+\r\n", f);
		}
	}
	if (meta->trace_header) {
		metalist_t *t;
		t=meta->trace_header;
		while (t) {
			unsigned long n=strtol(t->text,NULL,10);
			if (n & opts.debugmask_maustausch) {
				const char *nx;
				if (n & MMT_ERROR)
					nx="Error: ";
				else
					nx="";
				fprintf (f, "!%s%s" LINEEND, nx,t->ext);
			}
			t=t->next;
		}
	}

	/* ich hoffe an dieser Stelle irgendwann >-Zeilen ausgeben zu drfen */

	/* So, nun sind Textkludges auszugeben */

	/* Eventuell pate der Content-Type nicht in die Headerzeile */
	if (m_content_type_for_body) {
		textsize += fprintf (f, ":%s" LINEEND, m_content_type_for_body);
		free (m_content_type_for_body);
	}
	/* sonstige MIME-relevante Header ausgeben */
	if (meta->content_id)
		textsize += fprintf (f, ":Content-ID: %s" LINEEND, meta->content_id);
	if (meta->content_description)
		textsize += fprintf (f, ":Content-Description: %s" LINEEND,
							 meta->content_description);

	if (meta->reply_to && !opts.send_reply_to) {
		/* nur *ein* Reply-To, mehr ist wohl nicht sinnvoll, oder? */
		textsize += fprintf (f, ":Reply-To: %s" LINEEND, meta->reply_to->text);
		used_cludges = 1;
	}
	if (meta->sender && !opts.send_sender) {
		textsize += fprintf (f, ":Sender: %s" LINEEND, meta->sender->text);
		used_cludges = 1;
	}
	if (meta->keywords) {
		textsize += fprintf (f, ":Keywords: %s" LINEEND, meta->keywords);
		used_cludges = 1;
	}
	if (meta->summary) {
		textsize += fprintf (f, ":Summary: %s" LINEEND, meta->summary);
		used_cludges = 1;
	}
	if (is_crossposting) {
		metalist_t *ng;

		textsize += fprintf (f, ":Newsgroups: ");
		ng = meta->newsgroups;
		while (ng) {
			textsize += fprintf (f, "%s", ng->text);
			if (ng->next)
				textsize += fprintf (f, ",");
			else
				textsize += fprintf (f, LINEEND);
			ng = ng->next;
		}
		used_cludges++;
	}
	if (need_followup_to_kludge && meta->followup_to) {
		/* Bei News Followup-Kludge erzeugen */
		metalist_t *ng;

		textsize += fprintf (f, ":Followup-To: ");
		ng = meta->followup_to;
		while (ng) {
			if (!strcasecmp (ng->text, "poster"))
				textsize += fprintf (f, "poster");
			else
				textsize += fprintf (f, "%s", ng->text);
			if (ng->next)
				textsize += fprintf (f, ",");
			else
				textsize += fprintf (f, LINEEND);
			ng = ng->next;
		}
		used_cludges++;
	}
	if (header_envelope_mismatch) {
		metalist_t *l;
		size_t ll=0;
		ll += fprintf (f, ":X-To: ");
		for (l=meta->head_to;l && l->text; l=l->next) {
			if (l!=meta->head_to && strlen(l->text)+ll >77) {
				textsize+=ll+3;
				fprintf (f, "," LINEEND ":  ");
				ll=2; /* zwei spaces */
			} else if (l!=meta->head_to) {
				ll+=fprintf (f, ", ");
			}
			ll += fprintf (f, "%s",l->text);
		}
		ll+=fprintf (f, LINEEND);
		textsize+=ll;
	}
	if (is_mime) {
		textsize += fprintf (f, ":-" LINEEND);
	} else if (used_cludges) {
		/* hier sollte nichts mehr ankommen! -> MIME */
		/* Dann eine Leerzeile nach den Pseudoheadern einbauen */
		textsize += fprintf (f, ":%s", LINEEND);
	}
	return textsize;
}

/* 
 * Die zentrale Exportroutine fr Nachrichten im MausTausch-Format.
 * Sie wird durch die Notwendigkeit, lngere Nachrichten zu splitten,
 * nicht unbedingt eleganter. 
 */
static mio_retcode_t
mt_export (mio_t * para, metamsg_t * meta)
{
	struct mt_private *priv = (struct mt_private *) para->priv_data;
	FILE *f;
	char *start;
	int partnr;

	assert (priv);
	assert (meta);

	f = priv->f;

	start = meta->body.data;
	partnr = 0;

	while (start && *start) {
		size_t hlen;
		int llen;

		hlen = export_part (f, meta, partnr);
		/* Zuerst mal die relativ schne Variante ohne TAB-Wandelung. */
		if (!opts.tabsize) {
			long in_line = 0;
			while (start && *start) {
				char *p;
				p = strpbrk (start, "\r\n");
				if (p)
					llen = p - start;
				else
					llen = strlen (start);

				/* Immer eine Zeile durchlassen */
				if (in_line++ > 0 && opts.split_after
					&& llen + 2 + hlen > ((size_t) opts.split_after))
					break;
				putc (':', f);
				fwrite(start,llen,1,f);
				fwrite(LINEEND,2,1,f);
				hlen += llen + 2;
				if (!p)
					start = NULL;
				else {
					if (*p == '\r')
						p++;
					if (*p == '\n')
						p++;
					start = p;
				}
			}
		} else {				/* Variante mit TAB-Wandelung */
			int in_line = 0;
			while (start && *start) {
				char *crlf;
				char *q;
				size_t lc = 0;


				crlf = strchr (start, '\n');
				if (crlf && crlf > start && crlf[-1] == '\r')
					crlf--;		/* auf CR */
				for (q = start; (crlf) ? q < crlf : *q; q++) {
					lc++;
					if (*q == '\t')
						lc += opts.tabsize - (lc % opts.tabsize);
				}
				/* lc enthaelt nun die Laenge des Ausgabestrings ohne \r\n */
#if 0
				fprintf (f, "%ld + %ld = %ld\n",
						 (long) lc, (long) hlen, (long) lc + hlen + 2);
#endif
				/* Immer eine Zeile durchlassen */
				if (in_line++ > 0 && opts.split_after
					&& lc + hlen + 2 > (size_t) opts.split_after)
					break;
				hlen += lc + 2;
				putc (':', f);

				lc = 0;
				for (q = start; (crlf) ? q < crlf : *q; q++) {
					if (*q == '\t') {
						do {
							putc (' ', f);
							lc++;
						}
						while (lc % opts.tabsize);
					} else {
						putc (*q, f);
						lc++;
					}
				}
				fputs (LINEEND, f);
				if (!crlf)
					start = NULL;
				else {
					if (*crlf == '\r')
						start = crlf + 2;
					else
						start = crlf + 1;
				}
			}
		}
		/* Falls die Nachricht gesplittet werden mute, dann die
		 * Distribution setzen, um Dupes zu verhindern.
		 */
		if (!meta->is_mail) {
			if (opts.workmode == GATE && (partnr || (start && *start))) {
				fprintf (f, "DL" LINEEND);
			} else if (meta->distribution) {
				if (!strcasecmp (meta->distribution, opts.usenet_dist_mausnet))
					fprintf (f, "DM" LINEEND);
				else if (!strcasecmp (meta->distribution, opts.pollsite)
						 || !strcasecmp (meta->distribution, opts.usenet_dist_lokal))
					fprintf (f, "DL" LINEEND);
				else
					fprintf (f, "DN" LINEEND);
			}
		}
		partnr++;
	}							/* while noch etwas brig */
	if (meta->is_mail && opts.generate_empty_mail_body) {
		/* PMs mit wichtigem Betreff ohne Text *knnen* auftreten */
		if (!meta->body.data || !*meta->body.data) {
			export_part (f, meta, partnr);
			fprintf (f, ":[kein Text vorhanden -- mauscon]" LINEEND);
		}
	}
	if (fflush (f) == 0) {
		msglog ("# %s: -> maustausch", meta->message_id);
		return MIO_OK;
	} else {
		msglog ("# %s: -> %s: %s", meta->message_id,
				_ ("maustauschexport gescheitert"), strerror (errno));
		return MIO_ERROR;
	}
}

static void
make_path (metamsg_t * m, const char *group_path_entry,
		   const char *foreign_export_path, const char *abs_unk_path)
{
	/* Eine Path-Zeile zusammensetzen.
	 * Gegenueber frueheren MausConnectversionen stark vereinfacht. 
	 * Den Pfad einer Nachricht im MausNet zu rekonstruieren ist 
	 * ziemlich witzlos.
	 */

	char *n;

	/* Den Namen unserer Kiste und den der Pollmaus in den Pfad setzen. */
	n = concat (
		(abs_unk_path && *abs_unk_path) ? abs_unk_path : "",
		(abs_unk_path && *abs_unk_path) ? "!" : "",
		(foreign_export_path && *foreign_export_path) 
				? foreign_export_path : "",
		(foreign_export_path && *foreign_export_path) ? "!" : "",
		(group_path_entry && *group_path_entry) ? group_path_entry : "",
		(group_path_entry && *group_path_entry) ? "!" : "",
		(opts.path_entry && *opts.path_entry) ? opts.path_entry : "",
		(opts.path_entry && *opts.path_entry) ? "!" : "",
		opts.pollsitedomain,
		"!not-for-mail",
		NULL);
	XNULLCHECK (n);
	m->path = entumlauten1 (n); /* just to be save */
	free(n);
	XNULLCHECK (m->path);
	if (foreign_export_path && *foreign_export_path) {
		meta_add_trace (m,MMT_PATH, "add %s to path, foreign export", foreign_export_path);
	}
	if (group_path_entry && *group_path_entry) {
		meta_add_trace (m,MMT_PATH, "add %s to path, due to newsgroup config", group_path_entry);
	}
}

static void
mt_parse_x_line(metamsg_t *msg, const char *s)
{
	while (*s)
	{
		if (!s[1])
			break;
		switch(*s)
		{
		case 'R': 
			if (s[1]=='-')  {
				msg->mausflags.exportflag=XR_MINUS; 
				meta_add_trace(msg,MMT_NX,"noexport set, XR- found");
			} else if (s[1]=='+') {
				msg->mausflags.exportflag=XR_PLUS; 
			}
			break;
		}
		s+=2;
	}
}


static mio_retcode_t
mt_import (mio_t * channel, metamsg_t * meta)
{
	struct mt_private *priv;

	int ende_bei_doppelkreuz;
	int in_text_kludges = 0;	/* In header or pseudoheader? */
	int raus = 0;
	mio_retcode_t retcode = MIO_OK;
	char *line;
	size_t llen;
	unsigned int unknown_group_count = 0;
	grp_data_t *one_unknown_group = NULL;

	/* these are static to avoid too many malloc/free calls. 
	 * Remember: we go through this 1000 to 2000 times */
	static uostr_t zwi_fromrealname=UOSTR_INIT;
	static uostr_t zwi_fromadress=UOSTR_INIT;
	static uostr_t tmp_ad=UOSTR_INIT;
	static uostr_t tmp_rn=UOSTR_INIT;
	static uostr_t zwi_date=UOSTR_INIT;	/* E */
	static uostr_t zwi_to=UOSTR_INIT;	/* A */
	static uostr_t zwi_maus_ref=UOSTR_INIT;	/* - */
	static uostr_t zwi_real_ref=UOSTR_INIT;	/* R */
	static uostr_t zwi_real_id=UOSTR_INIT;	/* I */
	static uostr_t zwi_maus_id=UOSTR_INIT;	/* # */

	const char *group_path_entry = NULL;
	const char *foreign_export_path = NULL;
	BOOLEAN no_export = FALSE;
	BOOLEAN ignore_all_headers=FALSE; /* beim "entsplitten" benutzt */
	BOOLEAN got_grp_allow_transit=FALSE;
	BOOLEAN got_grp_disallow_transit=FALSE;
	BOOLEAN allow_transit=FALSE;
	BOOLEAN last_kludge_unknown=FALSE;
	BOOLEAN force_noexport=FALSE;

	int b_found = 0;
	int groups_found = 0;
	int m_found = 0;

	char *absender_unkonvertierbar = NULL;
	const char *abs_unk_path = NULL;
	const char *maybe_organization=NULL;

	assert (channel);
	assert (meta);

	uostr_fcut(&zwi_fromrealname);
	uostr_fcut(&zwi_fromadress);
	uostr_fcut(&tmp_ad);
	uostr_fcut(&tmp_rn);
	uostr_fcut(&zwi_date);
	uostr_fcut(&zwi_to);
	uostr_fcut(&zwi_maus_ref);
	uostr_fcut(&zwi_real_ref);
	uostr_fcut(&zwi_real_id);
	uostr_fcut(&zwi_maus_id);

	priv = (struct mt_private *) channel->priv_data;

	if (priv->eof)
		return MIO_EOF;


	/* Erst mal die Nachricht auf einen definierten leeren Zustand setzen */
	meta_clean_message (meta);

	ende_bei_doppelkreuz = 0;

	if (priv->zurueckgestellt) {
		line = priv->line;
		llen = priv->llen;
		priv->zurueckgestellt=FALSE;
		goto do_this_line; /* keep it out of the main loop */
	}
	/* Zeilenweise lesen */
	while (TRUE) {
		{
			ssize_t got;
			char *crlfp;
			got=uoio_getdelim_zc(&priv->uoio,&line,'\n');
			if (got<0) {
				log(LOG_WARNING,"%s: %s",priv->fname,strerror(errno));
				retcode = MIO_ERROR;	/* Lesefehler */
				llen=0; /* keep gcc quiet */
				break;
			} else if (got==0) {
				if (!ende_bei_doppelkreuz)
					retcode = MIO_EOF;	/* Ende des Files */
				llen=0; /* keep gcc quiet */
				break;
			} else {
				llen=(size_t) got;
			}
			/* strpbrk ist verhltnismig teuer */
			crlfp = llen + line - 1;
			if (*crlfp != '\n') {
				char *p;
				static char *my_leak=NULL;
				log(LOG_WARNING,"last line ended without '\\n'");
				/* hack around it */
				if (my_leak)
					free(my_leak);
				my_leak=p=XMALLOC(llen+1);
				memcpy(p,line,llen);
				line=p;
				line[llen]=0;
			} else {
				*crlfp-- = 0;
				llen--;
				if (*crlfp == '\r') {
					*crlfp-- = 0;
					llen--;
				}
			}
		}
	  do_this_line:

		if (ignore_all_headers && line[0]!=':' && line[0]!='#') {
			continue;
		}
		if (line[0]!='#' && line[0]!=':' && line[0]!='>') {
			/* leere Headerzeilen? */
			if (!line[1]) {
				force_noexport=TRUE;
				meta_add_trace (meta,MMT_MIME|MMT_NX,
							   "found empty %c-Line",line[0]);
				continue;
			}
		}
		switch (line[0]) {
		case '#':
			if (line[1] == '\0') {
				/* Eine Zeile mit nur '#' beendet das Outfile */
				priv->eof = TRUE;
				/* das passiert, wenn die erste Zeile gleich "#" ist */
				if (!zwi_maus_id.len && !zwi_real_id.len)
					retcode = MIO_EOF;
				else
					retcode = MIO_OK;
				raus = 1;
				break;
			}
			/* prfen auf Fortsetzungsteil gesplitteter Nachrichten */
			if (meta->maus_id) {
				char *at;
				at = strchr (line + 1, '@');
				if (at) {
					char *colon = strchr (at + 1, ':');
					if (colon) {
						if (strncmp (line + 1, meta->maus_id, colon - line - 1) == 0) {
							/* ok, Fortsetzung *dieser* Nachricht */
							ignore_all_headers = TRUE;
							break;	/* out of case */
						}
					}
				}
			}
			if (ende_bei_doppelkreuz) {
				priv->zurueckgestellt = TRUE;
				priv->line=line;
				priv->llen=llen;
				retcode = MIO_OK;	/* noch nicht zuende */
				raus = 1;
				break;
			}
			ende_bei_doppelkreuz = 1;


			if (strchr (line + 1, '@')) {
				/* Muss sein. Boxkrzel knnen Umlaute enthalten */
				xentumlauten2uostr_dup(line+1,&zwi_maus_id);
				uostr_x0(&zwi_maus_id);
				meta->maus_id = XSTRDUP (zwi_maus_id.data);
			} else {
				/* Spezialnachricht */
				unsigned char timezwi[STR_ULONG];
				minfo_t *found;

				uo_datetime_t dt;
				uo_sec70_t t;
				t=uo_now();
				uo_sec702dt(&dt,&t);

				str_ulong(timezwi,(unsigned long)t);
				/* ID erzeugen */
				uostr_xdup_cstrmulti(&zwi_real_id,line+1,timezwi,"@",opts.mysitedomain,NULL);
				meta->date=uo_now();
				uostr_x0(&zwi_real_id);
				/* Special_ID sichern */
				meta->maus_special_id = XSTRDUP (line + 1);

				yyyy_mm_dd_hh_mm_ss(timezwi,&dt,'-');

				/* Sender erzeugen */
				uostr_cut(&tmp_ad,0);
				uostr_xadd_cstrmulti(&tmp_ad,"MausConnect@",opts.mysitedomain,NULL);
				uostr_x0(&tmp_ad);
				/* Fehler ist nicht so wild */
				meta_add_list (&meta->sender, tmp_ad.data, NULL, 0);

				/* "From" erzeugen */
				uostr_xdup_cstrmulti(&zwi_fromrealname,opts.boxtypname, " ",
						 opts.pollsite,NULL);
				uostr_x0(&zwi_fromrealname);
				uostr_xdup_cstrmulti(&zwi_fromadress, opts.boxtypname, "@",
						 opts.pollsitedomain,NULL);
				uostr_x0(&zwi_fromadress);
				str_lwr(zwi_fromadress.data);

				found = info_find (line + 1);
				if (found) {
					/* Ein Infofile */

					/* "Subject" */
					if (!found->kurztext || !*(found->kurztext)) {
						if (line[1] == 'I')
							meta->subject = xconcat("MAUSINFO ", line+1, " (", timezwi,")",NULL);
						else
							meta->subject = xconcat("MAUSDATEN ", line+1, " (", timezwi,")",NULL);
					} else {
						meta->subject = xconcat(found->kurztext, " (", timezwi,")",NULL);
					}
				} else if (!strcmp (line + 1, "LOG")) {
					/* "Subject" */
					meta->subject = xconcat("LOG ", timezwi,NULL);
				} else if (!strcmp (line, "#HEAD")) {
					/* Eigentlich berflssig, wird weiter oben in der
					 * Hierarchie sowieso verworfen. 
					 */
					meta->subject = xconcat("HEAD ", timezwi,NULL);
				} else if (!strcmp (line, "#REN")) {
					/* Eigentlich berflssig, wird weiter oben in der
					 * Hierarchie sowieso verworfen. 
					 * oder wird es mittlerweile ausgewertet? Hmm - ja?
					 */
					meta->subject = xconcat("REN ", timezwi,NULL);
				}
				/* unbekannte Bloecke */
				else {
					meta->subject = xconcat("MAUSDATEN ", line+1, " (", timezwi,")",NULL);
				}
			}
			break;
		case '-':
			xentumlauten2uostr_dup(line+1,&zwi_maus_ref);
			uostr_x0(&zwi_maus_ref);
			break;
		case 'A':				/* an */
			uostr_xdup_cstr(&zwi_to,line+1);
			uostr_x0(&zwi_to);
			break;
		case 'a':				/* postfach */
			meta->quark_a_zeile = XSTRDUP (line + 1); /* XXX moegliches Speicherleck */
			break;
		case 'B':
			b_found = 1;
			meta->maus_status = XSTRDUP (line + 1); /* XXX moegliches Speicherleck */
			break;
		case 'D':
			if (line[1] == 'L')
				meta->distribution = XSTRDUP (opts.usenet_dist_lokal);
			else if (line[1] == 'M')
				meta->distribution = XSTRDUP (opts.usenet_dist_mausnet);
			else if (line[1] != 'N') {
				static int dist_warned=0;
				if (dist_warned++==0) 
					log (LOG_INFO, "unknown MausTausch-distribution %c",line[1]);
				force_noexport=TRUE;
				meta_add_trace (meta,MMT_DIST|MMT_NX, 
								"noexport flag set, unknown distribution");
				meta_add_list (&meta->unknown_headers,
							   "X-Original-MAUS-D-Line",
							   line + 1, 0);
			}
			break;
		case 'E':
			maus_get_date(&meta->date,line+1);
			/* Fr Erzeugung der Message-ID sichern */
			if (!strchr (line, ' ') && llen - 1 >= 12) {
				uostr_xdup_cstr(&zwi_date,line+1);
				uostr_x0(&zwi_date);
			}
			break;
		case 'G':				/* gruppe */
			{
				grp_data_t *grp;

				grp = grp_suche_mausname (line + 1, opts.pollsite);
				if (!grp) {
					/* neue Gruppe, noch nicht in der Liste */
					uostr_t tmp=UOSTR_INIT;
					grp_create_rfcname (line + 1, &tmp);
					one_unknown_group = XMALLOC (sizeof (grp_data_t));
					*one_unknown_group = leer_grp;
					one_unknown_group->mausname = XSTRDUP (line + 1);
					one_unknown_group->rfcname = XSTRDUP (tmp.data);
					one_unknown_group->boxen = XSTRDUP (opts.pollsite);
					grp = one_unknown_group;
					unknown_group_count++;
					uostr_freedata(&tmp);
				}
				if (grp->allow_transit==TRI_TRUE)
					got_grp_allow_transit=TRUE;
				else if (grp->allow_transit==TRI_FALSE)
					got_grp_disallow_transit=TRUE;
				no_export |= grp->no_export;
				if (!group_path_entry)
					group_path_entry = grp->path_entry;
				/* hinten an die Liste der Gruppen anhaengen */
				if (!meta_add_list (&meta->newsgroups, grp->rfcname, NULL, 0)) {
					log (LOG_CRIT, "out of memory");
					exit (1);
				}
				groups_found++;
			}
			break;
		case 'F': /* Followup-To */
			if (strchr(line+1,' ')) {
				log (LOG_DEBUG, 
					_ ("illegal F(followup-to) line, contains whitespace %s"), 
						line + 1);
				meta_add_trace (meta,MMT_ERROR|MMT_NX,
							   "found illegal F (Followup-To) line");
				meta_add_list (&meta->unknown_headers,
							   "X-Original-MAUS-F-Line",
							   line + 1, 0);
			} else if (!strcasecmp (line +1 , "PM")) {
				meta_add_trace(meta,MMT_INFO,
				   "converted %s to %s", line,"poster");
				if (!meta_add_list (&meta->followup_to, "poster", NULL, 0)) {
					log (LOG_CRIT, _ ("out of memory"));
					exit (1);
				}
			} else {
				grp_data_t *grp;
				grp = grp_suche_mausname (line + 1, opts.pollsite);
				if (!grp) {
					/* neue Gruppe, noch nicht in der Liste. 
					 * wird nicht abgespeichert!
					 */
					uostr_t gtmp=UOSTR_INIT;
					grp_data_t leer = LEER_GRP;
					grp_create_rfcname (line + 1, &gtmp);
					grp = XMALLOC (sizeof (grp_data_t));
					*grp = leer;
					grp->mausname = XSTRDUP (line+1);
					grp->rfcname = XSTRDUP (gtmp.data);
					grp->boxen = XSTRDUP (opts.pollsite);
					grp->dont_save = 1;
					grp_append_new (grp);
					uostr_freedata(&gtmp);
				}
				if (!meta_add_list (&meta->followup_to, grp->rfcname, NULL, 0)) {
					log (LOG_CRIT, _ ("out of memory"));
					exit (1);
				}
			}
			break;
		case 'I':
			/* Entumlauten: MAUS hat soetwas schon mal geliefert */
			xentumlauten2uostr_dup(line+1,&zwi_real_id);
			uostr_x0(&zwi_real_id);
			break;
		case 'M':
			/* MIME-Versionen != 1.0 knnen wir nicht. 
			 * Gibt es jetzt auch nicht.
			 */
			if (0 == strncmp (line + 1, "1.0;", 4)) {
				char *p;
				meta->mime_version = XSTRDUP ("1.0");
				p = line + 5;
				while (isspace (*p))
					p++;
				if (0 == strncasecmp ("Content-Type:", p, 13)) {
					p += 13;
					while (isspace (*p))
						p++;
					meta->content_type = XSTRDUP (p);
				}
				m_found = 1;
			} else {
				force_noexport=TRUE;
				meta_add_trace (meta,MMT_MIME|MMT_NX,
							   "found unknown Mime-Version");
				meta_add_list (&meta->unknown_headers,
							   "X-Original-MAUS-M-Line",
							   line + 1, 0);
			}
			/* in jedem Fall haben wir jetzt "standardisierte" 
			 * Textkludges auszuwerten */
			in_text_kludges = 1;
			break;
		case 'N':
			uostr_xdup_cstr(&zwi_fromrealname,line+1);
			uostr_x0(&zwi_fromrealname);
			break;
		case 'O':
			meta->organization = strdup (line + 1);
			XNULLCHECK (meta->organization);
			break;
		case 'R':
			xentumlauten2uostr_dup(line+1,&zwi_real_ref);
			uostr_x0(&zwi_real_ref);
			break;
		case 'V':				/* Von */
			if (!meta->organization) {
				box_t *box;
				char *p;
				p = strchr (line + 1, '@');
				if (p) {
					p++;
					while (isspace (*p))
						p++;
					box = dom_find_boxkuerzel (p);
					if (!box)
						box = dom_find_domain (p);
					if (box) {
						maybe_organization=box->orga;
					}
				}
			}
			if (!mausadr_to_rfcadr (line + 1, &zwi_fromadress,
						zwi_fromrealname.len ? &tmp_rn : &zwi_fromrealname)) {
				log (LOG_WARNING, _ ("cannot convert %s"), line + 1);
				meta_add_trace (meta,MMT_ERROR|MMT_NX, "found illegal V (From) line");
				meta_add_list (&meta->unknown_headers,
							   "X-Original-From",
							   line + 1, 0);
				/* das setzt auch XR- */
				absender_unkonvertierbar = XSTRDUP(line+1);
			}
			break;
		case 'W':
			meta->subject = XSTRDUP (line + 1);
			break;
		case 'X':
			mt_parse_x_line(meta,line+1);
			meta_add_list (&meta->unknown_headers,
						   "X-MAUS-X-Line",
						   line + 1, 0);
			break;
		case 'Y':
			meta->x_gateway = XSTRDUP (line + 1);
			break;
		case 'T':
			{
				if (!mausadr_to_rfcadr (line+1, &tmp_ad, &tmp_rn)) {
					meta_add_trace (meta,MMT_ERROR,
								   "found illegal T (Reply-To) line", 0);
					meta_add_list (&meta->unknown_headers,
								   "X-Original-MAUS-T-Line",
								   line+1, 0);
				} else if (!meta_add_list (&meta->reply_to, tmp_ad.data, 
					tmp_rn.len ? tmp_rn.data : NULL, 0)) {
					log (LOG_CRIT, _ ("out of memory"));
					exit (1);
				}
			}
			break;
		case 'S':
			if (!mausadr_to_rfcadr (line+1, &tmp_ad, &tmp_rn)) {
				meta_add_trace (meta,MMT_ERROR,
							   "found illegal S (Sender) line", 0);
				meta_add_list (&meta->unknown_headers,
							   "X-Original-MAUS-S-Line",
							   line+1, 0);
			} else if (!meta_add_list (&meta->sender, tmp_ad.data, 
					tmp_rn.len ? tmp_rn.data : NULL, 0)) {
				log (LOG_CRIT, _ ("out of memory"));
				exit (1);
			}
			break;
		case ':':
			/* 
			 * Textzeilen? Irrtum - mit M-Zeile Textkludges.
			 */
			if (in_text_kludges) {
				char *inhalt;
				char *doppelpunkt;
				char *ws;
				if (!strcmp (line + 1, "-")) {
					/* Ende der Kludges erreicht */
					in_text_kludges = 0;
					goto text_verarbeitet;
				}
				/* nicht isspace - das ist locale-Abhngig */
				if (line[1]==' ' || line[1]=='\t') {
					/* Arg. Fortsetzungszeile. 
					 * Darf im MausNet nicht auftreten, aber wenn dieses Codestck
					 * erreicht wird dann _ist_ eine aufgetreten.
					 * Problem: Wir wissen nicht welche Zeile zuletzt da war.
					 */
					if (last_kludge_unknown) {
						/* ok, _hier_ geht es */
						meta_add_list (&meta->unknown_headers,
								   "\t", line+1, 0);
						meta_add_trace (meta,MMT_INFO,
									   "Fortsetzungszeile im erweiterten Header gefunden");
						goto text_verarbeitet;
					} else {
						char *skip=line+1;
						/* nicht isspace - das ist locale-Abhngig */
						while (*skip==' ' || *skip=='\t')
							skip++;
						meta_add_trace (meta,MMT_ERROR|MMT_NX,
									   "Fortsetzungszeile gefunden, nicht verarbeitet");
						meta_add_list (&meta->unknown_headers,
								   "X-Continuation",
								   skip, 0);
						last_kludge_unknown=TRUE;
						force_noexport=TRUE;
						goto text_verarbeitet;
					}
				} else 
					last_kludge_unknown=FALSE;

				doppelpunkt = strchr (line + 1, ':');
				if (!doppelpunkt
					|| ((ws = strpbrk (line + 1, "\t ")) != NULL && ws < doppelpunkt)) {
					in_text_kludges = 0;
					log(LOG_INFO,"found M-Line and illegal kludge: %s",line+1);
					meta_add_trace (meta,MMT_ERROR|MMT_NX,
								   "M-Line found, but illegal kludge line (first text line)");
					force_noexport=TRUE;
					goto normal_text;
				}
				inhalt = doppelpunkt;
				*inhalt++ = 0;
				while (isspace (*inhalt))
					inhalt++;
				if (!*inhalt) {
					/* Headerzeile ohne Inhalt? Hmmm ... */
					goto text_verarbeitet;
				}
				if (!strhalfcasecmp (line + 1, "content-type")) {
					meta->content_type = XSTRDUP (inhalt);
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "content-transfer-encoding")) {
					meta->content_transfer_encoding = XSTRDUP (inhalt);
					goto text_verarbeitet;
#if 0
	/* die sind besser in den unknown-Zeile aufgehoben, dann funktioniert auch 
	 * die Fortsetzungszeile ...
	 */
				} else if (!strhalfcasecmp (line + 1, "content-description")) {
					meta->content_description = XSTRDUP (inhalt);
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "content-id")) {
					meta->content_id = XSTRDUP (inhalt);
					goto text_verarbeitet;
#endif
				} else if (!strhalfcasecmp (line + 1, "followup-to")) {
					char *p2;
					uostr_t tmp=UOSTR_INIT;

					p2 = inhalt;
					while (*p2 && !isspace (*p2))
						p2++;
					if (strchr(p2,' ')) {
						log (LOG_DEBUG, 
							_ ("illegal Followup-To line, contains whitespace %s"), 
								inhalt);
						meta_add_trace (meta,MMT_ERROR,
									   "found illegal Followup-To");
						meta_add_list (&meta->unknown_headers,
									   "X-Original-MAUS-Followup-To",
									   inhalt, 0);
					} else if (!strcasecmp (inhalt, "PM")) {
						meta_add_trace(meta,MMT_INFO,
						   "converted %s to %s", inhalt,"poster");
						if (!meta_add_list (&meta->followup_to, "poster", NULL, 0)) {
							log (LOG_CRIT, _ ("out of memory"));
							exit (1);
						}
					} else {
						grp_data_t *grp;
						grp = grp_suche_mausname (inhalt, opts.pollsite);
						if (!grp) {
							/* neue Gruppe, noch nicht in der Liste. 
							 * wird nicht abgespeichert!
							 */
							grp_data_t leer = LEER_GRP;

							grp_create_rfcname (inhalt, &tmp);
							grp = XMALLOC (sizeof (grp_data_t));
							*grp = leer;
							grp->mausname = XSTRDUP (inhalt);
							grp->rfcname = XSTRDUP (tmp.data);
							grp->boxen = XSTRDUP (opts.pollsite);
							grp->dont_save = 1;
							grp_append_new (grp);
						} else
							uostr_xdup_cstr(&tmp,grp->rfcname);
					}
					if (!meta_add_list (&meta->followup_to, tmp.data, NULL, 0)) {
						log (LOG_CRIT, _ ("out of memory"));
						exit (1);
					}
					uostr_freedata(&tmp);
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "reply-to")) {
					if (!mausadr_to_rfcadr (inhalt, &tmp_ad, &tmp_rn)) {
						meta_add_trace (meta,MMT_ERROR,
									   "found illegal Reply-To", 0);
						meta_add_list (&meta->unknown_headers,
									   "X-Original-Reply-To",
									   inhalt, 0);
					} else if (!meta_add_list (&meta->reply_to, tmp_ad.data, 
						tmp_rn.len ? tmp_rn.data : NULL, 0)) {
						log (LOG_CRIT, _ ("out of memory"));
						exit (1);
					}
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "sender")) {
					if (!mausadr_to_rfcadr (inhalt, &tmp_ad, &tmp_rn)) {
						meta_add_trace (meta,MMT_ERROR,
									   "found illegal Sender", 0);
						meta_add_list (&meta->unknown_headers,
									   "X-Original-Sender",
									   inhalt, 0);
					} else if (!meta_add_list (&meta->sender, tmp_ad.data, 
						tmp_rn.len ? tmp_rn.data : NULL, 0)) {
						log (LOG_CRIT, _ ("out of memory"));
						exit (1);
					}
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "keywords")) {
					meta->keywords = strdup (inhalt);
					goto text_verarbeitet;
				} else if (!strhalfcasecmp (line + 1, "summary")) {
					meta->summary = strdup (inhalt);
					goto text_verarbeitet;
				}

				/* wenn wir hier ankommen haben wir eine unbekannte
				 * Kludgezeile. Nun gut, "shit happens".
				 */
				meta_add_list (&meta->unknown_headers,
							   line+1,
							   inhalt, 0);
				last_kludge_unknown=TRUE;

			} else {
			  normal_text:
			  	meta_add_to_body_crlf(meta,line+1,llen-1);
			}
		  text_verarbeitet:
			break;
		}						/* switch */
		if (raus)
			break;				/* JETZT - vor dem fgets! */
	}							/* while */

	if (retcode==MIO_EOF)
		return retcode;

	if (groups_found == 1 && one_unknown_group) {
		grp_append_new (one_unknown_group);
		log (LOG_INFO, _ ("new newsgroup %s -> %s"),
			 one_unknown_group->mausname, one_unknown_group->rfcname);
	}

	meta->is_mail = 0;
	if (b_found)
		meta->is_mail = 1;
	if (groups_found)
		meta->is_mail = 0;

	if (!meta->is_mail) {
		allow_transit=FALSE;
		if (got_grp_disallow_transit)
			allow_transit=FALSE;
		else if (got_grp_allow_transit)
			allow_transit=TRUE;
		else if (opts.allow_transit)
			allow_transit=TRUE;
	}


	/* Was muss jetzt noch geschehen? */

	/* Mglicherweise haben wir keine Organization, wohl aber 
	 * Infos aus der ITB 
	 */
	if (!meta->organization && maybe_organization)
		meta->organization = XSTRDUP (maybe_organization);

	/* Der Absender muss nun erzeugt werden */
	if (!meta_add_list (&meta->head_from, zwi_fromadress.data, 
		zwi_fromrealname.len ? zwi_fromrealname.data : NULL,
						0)) {
		log (LOG_CRIT, _ ("out of memory"));
		exit (1);
	}
	if (meta->is_mail)
		meta->env_from = XSTRDUP (zwi_fromadress.data);

	/* Empfaenger auswerten */
	if (zwi_to.len) {
		if (meta->newsgroups)
			meta->x_to = XSTRDUP (zwi_to.data);
		else {
			if (!mausadr_to_rfcadr (zwi_to.data, &tmp_ad, &tmp_rn)) {
				meta_add_trace (meta,MMT_ERROR, "found illegal A (to) line");
				meta_add_list (&meta->unknown_headers,
							   "X-Original-To",
							   zwi_to.data, 0);
			}
			if (!meta_add_list (&meta->head_to, tmp_ad.data, 
				tmp_rn.len ? tmp_rn.data : NULL, 0)) {
				log (LOG_CRIT, _ ("out of memory"));
				exit (1);
			}
		}
	}
	/* Bei PMs muss jetzt eine Envelope-Adresse erzeugt werden */
	if (meta->env_to == NULL
		&& meta->newsgroups == NULL
		&& meta->head_to) {
		if (opts.workmode == GATE)
			meta->env_to = strdup (meta->head_to->text);
		else
			meta->env_to = strdup (opts.mailuserdomain);
		XNULLCHECK (meta->env_to);
	}
	/* - Eine ID mu erzeugt werden. Die MAUS-internen IDs sind ja 
	 *   bekanntlich nicht eindeutig.
	 */
	if (zwi_real_id.len) {			/* I */
		meta->message_id = xentumlauten1 (zwi_real_id.data);
	} else if (zwi_maus_id.len
				&& (strchr (zwi_maus_id.data, '.')
			   || !(isupper (zwi_maus_id.data[0]) && isdigit (zwi_maus_id.data[1])))) {	/* # */
ist_lange_id:
		/* Auf irgendeinem Weg ist eine vollstndige ID in die # -Zeile
		 * gekommen. vielleicht ein use2maus oder Quark 2.x.
		 */
		meta->message_id = XSTRDUP (zwi_maus_id.data);
	} else if (zwi_date.len) {
		char *p;
		BOOLEAN ist_maus_krueppel_id=TRUE;

		/* die Tcke hierbei ist: wir knnten an einer Quark laufen (also lange
		 * ID in #-Zeile bzw. jetzt zwi_maus_id), und die Nachricht knnte eine
		 * ID dieser Art haben: E0x6Sdn-0006yG-00@wrath
		 * also ohne . hinter dem @. Vorsicht Dupegefahr.
		 * opts.boxtyp abzufragen ist verlockend, aber Betreiber machen schon
		 * mal Konfigurationsfehler, nicht wahr, Uwe?
		 * ##### FIXME: der Code hier weiss zuviel ber den Aufbau von Message-IDs!
		 */
		p=zwi_maus_id.data;
		if (*p!='A' && *p!='P')
			ist_maus_krueppel_id=FALSE;
		else {
			char *q=++p;
			while (isdigit(*p))
				p++;
			if (*p!='@') {
				ist_maus_krueppel_id=FALSE;
			} else {
				long n=strtol(q,NULL,10);
				if (n<=0 || n>65535) {
					ist_maus_krueppel_id=FALSE;
				} else {
					p++;
					/* isn't it ugly?. isalpha weil Kleinbuchstaben bei der 
					 * Entumlautung entstehen knnen.
					 */
					while (isalpha(*p)
						|| *p=='' || *p=='' || *p=='' || *p=='-')
							p++;
					if (*p)
						ist_maus_krueppel_id=FALSE;
					/* else knnten wir jetzt noch nachsehen ob hinter dem 
					 * @ ein bekanntes Boxkrzel steht. Aber erstens ist das
					 * teuer, und zweitens steht auch ab und zu schrott in 
					 * den kurzen MAUS-IDs (wenn man irgendeine Box mit 
					 * Umlauten im Krzel die Umlaute verhunzt hat. Passiert
					 * ein bis zweimal im Jahr).
					 */
				}
			}
		}
		if (!ist_maus_krueppel_id) {
			/* not too serious. */
			meta_add_trace (meta,MMT_ID, "found strange message-id in # line");
			goto ist_lange_id; /* it's sometimes ok to goto. */
		}
		meta_add_trace(meta,MMT_NX|MMT_ID,"hacked message-ID from short MAUS-ID");
		force_noexport=TRUE;

		{
			char *tmp=concat(zwi_date.data,".",zwi_maus_id.data,opts.mausdomain,NULL);
			XNULLCHECK (tmp);
			meta->message_id = entumlauten1 (tmp);
			XNULLCHECK (meta->message_id);
			free(tmp);
		}

		/* Selbst zusammengesetzte IDs nach lowercase wandeln */
		p = meta->message_id;
		while (*p) {
			*p = tolower (*p);
			p++;
		}
	} else {
		/* Schrott, Infofile */
		char *tmp=concat(zwi_maus_id.data,opts.mausdomain,NULL);
		XNULLCHECK(tmp);
		meta->message_id = entumlauten1 (tmp);
		XNULLCHECK (meta->message_id);
		free(tmp);
	}

	/* - Eine References mu erzeugt werden */
	/* beide Strings sind beim Einlesen schon entumlautet worden */
	if (zwi_real_ref.len) {
		meta_add_list (&meta->references, zwi_real_ref.data, NULL, 0);
	} else if (zwi_maus_ref.len && strchr (zwi_maus_ref.data, '.')) {
		meta_add_list (&meta->references, zwi_maus_ref.data, NULL, 0);
	}

	/* XR? ergaenzen */
	if (!meta->is_mail && !allow_transit) {
		/* Linkit erzeugt nie XRirgendwas ... Wieso darf soetwas eigentlich noch
		 * verwendet werden? 
		 */
		if (meta->mausflags.exportflag == XR_NOTSET) {
			char *copy = XSTRDUP (meta->message_id);
			str_lwr (copy);
			if (!strstr (copy, opts.mausdomain)) {
				meta_add_trace (meta,MMT_NX, 
								"noexport was undefined, now set, foreign export");
				meta->mausflags.exportflag=XR_MINUS;
			}
			free(copy);
		}
		/* XXX wird vielleicht noch ein Override fr einkommende XR+ gebraucht? */
	}

	if (absender_unkonvertierbar) {
		meta_add_trace (meta,MMT_ERROR|MMT_NX, "found illegal V (from) line");
		meta_add_list (&meta->unknown_headers,
					   "X-Original-From",absender_unkonvertierbar,0);
		force_noexport=TRUE;
	}

	if (meta->mausflags.exportflag != XR_MINUS) {
		if (force_noexport)
			meta->mausflags.exportflag=XR_MINUS;
	}
	/* bei ffentlichen Nachrichten und foreign export und _Gatemodus_ die 
	 * Nachricht u.U. wegwerfen */
	if (opts.workmode!=USER && !meta->is_mail && !meta->maus_special_id 
		&& meta->mausflags.exportflag == XR_MINUS) {
		BOOLEAN did_it=FALSE;
		if (opts.foreign_export_path) {
			foreign_export_path = opts.foreign_export_path;
			meta_add_trace (meta,MMT_NX|MMT_PATH, 
							"noexport flag set -> foreign export path");
			did_it=TRUE;
		}
		if (opts.foreign_export_distribution) {
			free (meta->distribution);
			meta->distribution = XSTRDUP (opts.foreign_export_distribution);
			meta_add_trace (meta,MMT_NX|MMT_DIST, 
							"noexport flag set -> distribution %s",
							meta->distribution);
			did_it=TRUE;
		}
		if (opts.foreign_export_kill || !did_it) {
			if (!opts.foreign_export_kill) {
				syslog(LOG_INFO,
					   "XR- found, but opts.foreign_... not set - killing message!");
			}
			msglog ("# %s: foreign export", meta->message_id);
			retcode = MIO_RETRY;
		}
	}

	/* X-Format-Zeile */
	meta->x_format = strdup ("MausTausch");		/* Error doesn't matter */

	/* Pfad zusammensetzen */
	make_path (meta, group_path_entry, foreign_export_path, abs_unk_path);

	if (!meta->is_mail) {
		/* 
		 * Versuchen, noch ein paar Daten zusammenzuklauben, den
		 * Gruppennamen zu wandeln.
		 */
		metalist_t *ng;
		BOOLEAN had_followup = FALSE;
		BOOLEAN explicit_mime_header_true = FALSE;
		BOOLEAN explicit_mime_header_false = FALSE;
		BOOLEAN explicit_mime_body_true = FALSE;
		BOOLEAN explicit_mime_body_false = FALSE;

		if (meta->followup_to)
			had_followup = TRUE;

		for (ng = meta->newsgroups; ng; ng = ng->next) {
			grp_data_t *grp;
			grp = grp_suche_rfcname (ng->text, opts.pollsite);
			if (grp) {
				if (grp->default_followup && !had_followup) {
					meta_add_list (&meta->followup_to,
								   grp->default_followup, NULL, 0);
				}
				if (grp->mime_headers == TRI_TRUE)
					explicit_mime_header_true = TRUE;
				else if (grp->mime_headers == TRI_FALSE)
					explicit_mime_header_false = TRUE;
				if (grp->mime_body == TRI_TRUE)
					explicit_mime_body_true = TRUE;
				else if (grp->mime_body == TRI_FALSE)
					explicit_mime_body_false = TRUE;
				if (grp->approve && !meta->approved)
					meta->approved = XSTRDUP (grp->approve);
			}
		}
		if (explicit_mime_header_true)
			meta->mime_code_headers = TRUE;
		else if (explicit_mime_header_false)
			meta->mime_code_headers = FALSE;
		else
			meta->mime_code_headers = opts.mime_coded_headers;
		if (explicit_mime_body_true)
			meta->mime_code_body = TRUE;
		else if (explicit_mime_body_false)
			meta->mime_code_body = FALSE;
		else
			meta->mime_code_body = opts.mime_coded_body;
	} else {
		meta->mime_code_headers = opts.mime_coded_headers;
		meta->mime_code_body = opts.mime_coded_body;
	}
	/* MIME-Headerzeilen-Balast loswerden falls mglich.
	 * Einige Nachrichten kommen mit MIME-Headern um im erweiterten Header
	 * weitere nicht-MIME-Header durchschleusen zu knnen.
	 * XXX FIXME: Das ignoriert "text/plain; charset=us-ascii", aber die kommen hier
	 * nicht so hufig an.
	 */
	if (meta->mime_version && !meta->content_type) {
		free(meta->mime_version);
		meta->mime_version=NULL;
	}
	if (meta->mime_version && 0==strcasecmp(meta->content_type,"text/plain")) {
		free(meta->content_type);
		free(meta->mime_version);
		meta->mime_version=NULL;
		meta->content_type=NULL;
	}

	if (!meta->is_mail && opts.workmode == GATE && no_export) {
		msglog ("# %s: not exported", meta->message_id);
		retcode = MIO_RETRY;
	}
	if (retcode == MIO_EOF || retcode==MIO_ERROR)
		priv->zurueckgestellt = FALSE;
	return retcode;
}


static void
maus_mk_ref (metamsg_t * msg, uostr_t *mref, uostr_t *rref)
{
	char *m;
	metalist_t *l;

	if (!msg->references) {
		uostr_x0(mref);
		uostr_x0(rref);
		return;
	}
	l = msg->references;
	while (l->next)
		l = l->next;

	uostr_xdup_cstr(rref,l->text);
	uostr_x0(rref);
	m=maus_long_to_short_id(l->text);
	if (m) {
		uostr_xdup_cstr(mref,m);
		uostr_x0(mref);
		free(m);
	}
}

#if 0
static void
add_ref (metalist_t ** r, char *s)
{
	char *p;

	while (s && *s) {
		p = strchr (s, ' ');
		if (p)
			*p = '\0';
		meta_add_list (r, s, NULL, 0);	/* Fehler ist nicht soo wild */
		if (p)
			*p = ' ';
		s = p;
		/* nur zur Sicherheit */
		while (s && *s == ' ')
			s++;
	}
}
#endif
