#include <unistd.h>
#include <errno.h>
#include <stdlib.h> /* getenv */
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include "uostr.h"
#include "uoio.h"
#include "str_ulong.h"
#include "mtalib.h"

struct priv {
	int goterrs;
	uoio_t mess;
	pid_t pid;
};
static mta_t * mta_sm_create P__((void));
static mta_ret_t mta_sm_write P__((mta_t *mta,const char *s, size_t len));
static mta_ret_t mta_sm_close P__ ((mta_t *mta));
static mta_ret_t mta_sm_abort P__ ((mta_t *mta));
static mta_ret_t mta_sm_open P__ ((mta_t **,const char *,const char *, const char *));

mta_t mta_sm={
	mta_sm_open,
	mta_sm_write,
	mta_sm_abort,
	mta_sm_close,
	0
};

static mta_t *
mta_sm_create(void)
{
	struct priv *pr;
	mta_t *m;
	pr=malloc(sizeof(struct priv));
	if (!pr) return 0;
	m=malloc(sizeof(mta_t));
	if (!m) {free(pr); return 0;}
	memcpy(m,&mta_sm,sizeof(mta_t));
	memset(pr,0,sizeof(*pr));
	m->priv=pr;
	return m;
}

#define DEFAULTSM "/usr/lib/sendmail"
static mta_ret_t
mta_sm_open(mta_t **m, const char *path,const char *from, const char *to)
{
	int pipe_mess[2];
	int e;
	char *copy_to;
	pid_t pid;
	mta_t *mta;

	mta=mta_sm_create(); if (!mta) {e=errno; goto cleanup_nothing; }
	if (pipe(pipe_mess) == -1) { e=errno;goto cleanup0; }
	copy_to=malloc(strlen(to)+1);
	if (!copy_to) { e=ENOMEM; goto cleanup3; }
	strcpy(copy_to,to);

	pid = fork();
	if (pid==-1) { e=errno; goto cleanup4; }
	else if (pid==0) {
		union { char *c;const char *d;} disq;
#define DISQUALIFY(x) (disq.d=x,disq.c)
		char **argv;
		int count;
		char *t;
		count=0;
		t=copy_to;
		while (t && *t) {
			char *lf=strchr(t,'\n');
			count++;
			if (lf) t=lf+1; 
			else t=0;
		}
		count+=4; /* av0, -f , from, 0 */
		argv=malloc(count*sizeof(char *));
		if (!argv) _exit(120);

		if (!path || 0==strcmp(path,"DEFAULT")) path=DEFAULTSM;
		argv[0]=DISQUALIFY(path);
		argv[1]=DISQUALIFY("-f");
		argv[2]=DISQUALIFY(from ? from : "<>");
		count=3;
		t=copy_to;
		while (t && *t) {
			char *lf=strchr(t,'\n');
			if (lf) *lf=0;
			argv[count++]=t;
			if (lf) t=lf+1; 
			else t=0;
		}
		argv[count]=0;
		close(pipe_mess[1]);

		if (-1==dup2(pipe_mess[0],0)) _exit(120);
		execv(DISQUALIFY(path),argv);
		_exit(120);
	}
	close(pipe_mess[0]);
	uoio_assign_w(&((struct priv *)mta->priv)->mess,pipe_mess[1],write,0);
	((struct priv *)mta->priv)->pid=pid;
	*m=mta;
	free(copy_to);
	return MTA_SUCCESS;
  cleanup4: free(copy_to);
  cleanup3: close(pipe_mess[0]); close(pipe_mess[1]); 
  cleanup0: 
    free(mta->priv);
	free(mta);
  cleanup_nothing:
	errno=e; 
	mta_errtxt=strerror(errno);
	mta_hcmssc="(#4.3.0)";
	return MTA_DEFER;
}

static mta_ret_t
mta_sm_write(mta_t *mta,const char *s, size_t len)
{
	if (-1==uoio_write_mem(&((struct priv *)mta->priv)->mess,s,len)) 
		((struct priv *)mta->priv)->goterrs++;
	if (((struct priv *)mta->priv)->goterrs) {
		mta_errtxt=strerror(errno); mta_hcmssc="(#4.3.0)";
		return MTA_DEFER;
	}
	return MTA_SUCCESS;
}

static mta_ret_t
mta_sm_abort(mta_t *mta)
{
	struct priv *priv=mta->priv;
	kill(priv->pid,SIGTERM);
	waitpid(priv->pid,0,0);
	close(priv->mess.fd);
	uoio_destroy(&priv->mess);
	free(priv);
	free(mta);
	return MTA_SUCCESS;
}

static mta_ret_t
mta_sm_close(mta_t *mta)
{
	int status;
	int exitcode;
	pid_t pid;
	struct priv *priv=mta->priv;

	if (priv->goterrs) { mta_errtxt="I/O error"; mta_hcmssc="(#4.3.0)"; }
	if (-1==uoio_flush(&priv->mess)) {priv->goterrs++;mta_errtxt=strerror(errno);mta_hcmssc="(#4.3.0)"; }
	if (priv->goterrs) {
		return mta_sm_abort(mta);
	}
	close(priv->mess.fd);
	uoio_destroy(&priv->mess);

	pid=priv->pid;
	free(priv);
	free(mta);

	if (waitpid(pid,&status,0)!=pid) { mta_errtxt="waitpid failed.";mta_hcmssc="(#4.3.0)"; return MTA_DEFER;}
	if (WIFSIGNALED(status)) {mta_errtxt="sendmail program crashed.";mta_hcmssc="(#4.3.0)"; return MTA_DEFER;};
	exitcode = WEXITSTATUS(status);

/* note: i'm using bernsteins hcmssc here for the same reason he uses them: DSN are almost useless,
 * and supporting them correctly ruins the code. I suppose djb thinks the same ...
 */
	switch (exitcode) {
	case 0: {
			static char pidbuf[20]; 
			str_ulong(pidbuf,pid);
			mta_errtxt=pidbuf; /* magic: on MTA_SUCCESS the ..._close() puts an id into the errtxt */
			return MTA_SUCCESS;
		}
		return MTA_DEFER;
	case 120:mta_errtxt="unable to execute sendmail program";mta_hcmssc="(#4.3.0)"; return MTA_DEFER;
	default:
		mta_errtxt="permanent problem";mta_hcmssc="(#5.3.0)"; 
		return MTA_FATAL;
	}
}
