/* syslog.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
/*
 * NAME
 *	syslog 3
 * SUMMARY
 *	surrogate BSD4.3 syslog facility
 * PROJECT
 *	pc-mail
 * PACKAGE
 *	nfs
 * SYNOPSIS
 *	include "syslog.h"
 *
 *	openlog(name, logopt, facility)
 *	char *name;
 *	int logopt;
 *	int facility;
 *
 *	syslog(priority, format, arguments)
 *	int priority;
 *	char *format;
 *
 *	closelog()
 * DESCRIPTION
 *	These functions emulate the BSD4.3 syslog(3) facility. Output is
 *	written to a system logfile (default /usr/spool/mqueue/syslog).
 *	That file should have mode 666 (i.e. read/write access permitted
 *	for everyone).
 *
 *	syslog() tries to output the log entry as one big chunk.
 * FILES
 *	/usr/spool/mqueue/syslog, system logfile
 * SEE ALSO
 *	percentm(3), interprets "%m" sequences in syslog format strings.
 * BUGS
 *	The functions use the stdio package. This may cause the program to
 *	grow unexpectedly.
 * AUTHOR(S)
 *	Wietse Z. Venema
 *	Eindhoven University of Technology
 *	Department of Mathematics and Computer Science
 *	Den Dolech 2, P.O. Box 513, 5600 MB Eindhoven, The Netherlands
 * CREATION DATE
 *	Sun Oct 29 15:12:57 MET 1989
 * LAST MODIFICATION
 *	10/29/89 22:58:33
 * VERSION/RELEASE
 *	1.2
 */

#include "config.h"
#ifndef HAVE_SYSLOG

#ifndef lint
static char sccsid[] = "@(#) syslog.c 1.2 10/29/89 22:58:33";

#endif

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#include <syslog.h>

extern int errno;

#ifndef SYSLOGFILE
#define	SYSLOGFILE	"syslog"
#endif

static char *logname;

/* percentm - replace %m by error message associated with value in err */

static char *
percentm(const char *str, int err)
{
    static char buf[BUFSIZ];
    const char *ip = str;
    char *op = buf;

    while (*ip) {
	switch (*ip) {
	case '%':
	    switch (ip[1]) {
	    case '\0':				/* don't fall off end */
			*op++ = *ip++;
			break;
	    case 'm':				/* replace %m */
			(void) strcpy(op,strerror(err));
			op += strlen(op);
			ip += 2;
			break;
	    default:				/* leave %<any> alone */
			*op++ = *ip++, *op++ = *ip++;
			break;
	    }
	default:
	    *op++ = *ip++;
	}
    }
    *op = '\0';
    return (buf);
}


/* openlog - initialize syslog facility; ignores all but the "name" argument */

/* ARGSUSED */

void openlog(const char *name, int logopt, int facility)
{
    logname = strdup(name);				/* ignore rest */
}

/* syslog - append entry to system log */

void
vsyslog(int pri, const char *fmt, va_list ap)
{
    long    secs;
    char   *date;
    FILE   *fp;
    int     err = errno;
    static char buf[BUFSIZ];

    if ((fp = fopen(SYSLOGFILE, "a"))==NULL) {
		return;
	}

	/* Format the time stamp */

	secs = time((long *) 0);
	date = asctime(localtime(&secs));
	(void) sprintf(buf, "%2.2s %3.3s %2.2s %8.8s %s: ",
		       date + 8, date + 4, date + 22, date + 11, 
				   logname ? logname : "unknown");

	/* Format the actual message */

	(void) vsprintf(buf + strlen(buf), percentm(fmt, err), ap);

	/* Try to output the log entry as one big chunk */

	(void) fprintf(fp, "%s\n", buf);
	(void) fclose(fp);
}

void
syslog(int pri, const char *fmt, ...)
{
    va_list ap;
	
    va_start(ap, fmt);
    vsyslog(pri, fmt, ap);
    va_end(ap);
}

/* closelog - a real dummy */

void closelog(void)
{
    /* no-op */
}

static int mask;
int setlogmask(int neu) 
{
	int old;
	old=mask;
	if (neu!=0)
		mask=neu;
	return old;
}

#endif					/* SYSLOG */
